import numpy as np

oo = np.iinfo(np.int32).max

# ------------------------------------------------------------------------------
# The Queue class (no error checking!)

class Queue:
    def __init__(self, size=21):
        self.queue = [None] * size
        self.size  = size
        self.head  = 1
        self.tail  = 1

    def Enqueue(self, item):
        self.queue[self.tail] = item
        if self.tail == self.size:
            self.tail = 1
        else:
            self.tail += 1

    def Dequeue(self):
        item = self.queue[self.head]
        if self.head == self.size:
            self.head = 1
        else:
            self.head += 1
        return item

    def NotEmpty(self):
        return self.head != self.tail

# ------------------------------------------------------------------------------
# The Vertex class

class Vertex:
    def __init__(self, label, color='WHITE', predec=None):
        self.label = label
        self.color = color
        self.dist = oo
        self.predec = predec

# ------------------------------------------------------------------------------
# Print shortest path

def PrintPath(s, v):
    if v.label == s.label:
        print(f'{s.label}', end=' ')
    elif v.predec == None:
        print(f'No path from {s.label} to {v.label}') 
    else:
        PrintPath(s, v.predec)
        print(f'{v.label}', end=' ')

# ------------------------------------------------------------------------------
# The input graph

graph = { 'r': ['s', 't', 'w'],
          's': ['r', 'u', 'v'],
          't': ['r', 'u'],
          'u': ['s', 't', 'y'],
          'v': ['s', 'w', 'y'],
          'w': ['r', 'v', 'x', 'z'],
          'x': ['w', 'y', 'z'],
          'y': ['u', 'v', 'x'],
          'z': ['w', 'x']
        }

vdict = {}
for vertex in graph.keys(): vdict[vertex] = Vertex(vertex)

# ------------------------------------------------------------------------------
# Starting vertex is vertex 's'. Set color to GRAY and distance to zero

s = vdict['s']
s.color = 'GRAY'
s.dist = 0

# ------------------------------------------------------------------------------
# Initialize the priority queue and enqueue the source vertex

Q = Queue()
Q.Enqueue('s')

# ------------------------------------------------------------------------------
# Examine all vertices in Q

while Q.NotEmpty():
    u = Q.Dequeue()
    uvert = vdict[u]
    for nbr in graph[u]:
        v = vdict[nbr]
        if v.color == 'WHITE':
            v.color = 'GRAY'
            v.dist = uvert.dist + 1 
            v.predec = uvert
            Q.Enqueue(nbr)
    uvert.color = 'BLACK'

# ------------------------------------------------------------------------------
# Print distances

for v in vdict.keys():
    vert = vdict[v]
    print(f'Vertex: {vert.label}  Distance: {vert.dist}')

# ------------------------------------------------------------------------------
# Print path from s to z

PrintPath(vdict['s'], vdict['z'])
print()


